#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/NGLInit.h"
#include "ngl/ShaderManager.h"
#include "ngl/Util.h"
#include "ngl/VertexArrayObject.h"
#include <ctime>


const static float INCREMENT=0.01;
const static float ZOOM=0.1;

//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

	// set this widget to have the initial keyboard focus
	setFocus();
	// re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
	this->resize(_parent->size());
	// Now set the initial GLWindow attributes to default values
	// Roate is false
	m_rotate=false;
	m_translate=false;
	// mouse rotation values set to 0
	m_spinXFace=0;
	m_spinYFace=0;
	m_origX=0;
	m_origY=0;
	m_modelPos.set(0,0,0);
	m_origXPos=0;
	m_origYPos=0;
}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
	// now we have finished remove the VOA
	m_vaoMesh->removeVOA();
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
	ngl::NGLInit *Init = ngl::NGLInit::instance();
	#ifdef WIN32
		glewInit();
	#endif
	Init->initGlew();

	glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
	// enable depth testing for drawing
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_MULTISAMPLE);
	// Now we will create a basic Camera from the graphics library
	// This is a static camera so it only needs to be set once
	// First create Values for the camera position
	ngl::Vector from(0,0,0.5);
	ngl::Vector to(0,0,0);
	ngl::Vector up(0,1,0);
	// create our camera
	m_cam= new ngl::Camera(from,to,up,ngl::PERSPECTIVE);
	// set the shape using FOV 45 Aspect Ratio based on Width and Height
	// The final two are near and far clipping planes of 0.5 and 10
	m_cam->setShape(25,(float)720.0/576.0,0.001,150,ngl::PERSPECTIVE);
	// now to load the shader and set the values
	// grab an instance of shader manager
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	// load a frag and vert shaders

	shader->createShaderProgram("TextureShader");

	shader->attachShader("SimpleVertex",ngl::VERTEX);
	shader->attachShader("SimpleFragment",ngl::FRAGMENT);
	shader->loadShaderSource("SimpleVertex","shaders/Vertex.vs");
	shader->loadShaderSource("SimpleFragment","shaders/Fragment.fs");

	shader->compileShader("SimpleVertex");
	shader->compileShader("SimpleFragment");
	shader->attachShaderToProgram("TextureShader","SimpleVertex");
	shader->attachShaderToProgram("TextureShader","SimpleFragment");
	// bind our attributes for the vertex shader
	shader->bindAttribute("TextureShader",0,"inVert");
	shader->bindAttribute("TextureShader",1,"inUV");
	shader->bindAttribute("TextureShader",2,"inNormal");
	shader->bindAttribute("TextureShader",3,"inTangent");
	shader->bindAttribute("TextureShader",4,"inBinormal");

	// link the shader no attributes are bound
	shader->linkProgramObject("TextureShader");
	(*shader)["TextureShader"]->use();
	// now pass the modelView and projection values to the shader
	shader->setShaderParamFromMatrix("TextureShader","ViewMatrix",m_cam->getModelView());
	shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
	// set our samplers for each of the textures this will correspond to the
	// multitexture id below
	shader->setShaderParam1i("TextureShader","tex",0);
  shader->setShaderParam1i("TextureShader","spec",1);
  shader->setShaderParam1i("TextureShader","normalMap",2);
	// number of lights we are using (loops in the shader)
	shader->setShaderParam1i("TextureShader","numLights",3);
	// specular power
	shader->setShaderParam1f("TextureShader","specPower",50.0);

	// build our VertexArrayObject from the mesh
	loadModel();

	// load and set a texture for the colour
	ngl::Texture t("textures/TrollColour.tiff");
	t.setMultiTexture(0);
	t.setTextureGL();
	// now one for the specular map
	ngl::Texture spec("textures/2k_troll_spec_map.jpg");
	spec.setMultiTexture(1);
	spec.setTextureGL();
	// this is our normal map
	ngl::Texture normal("textures/2k_ct_normal.tif");
	normal.setMultiTexture(2);
	normal.setTextureGL();

	// mip map the textures
	glGenerateMipmap(GL_TEXTURE_2D);
	glEnable(GL_MULTISAMPLE);
	/// now setup a basic 3 point lighting system
	ngl::Light key(ngl::Vector(-2,1,1),ngl::Colour(1,1,1,1),ngl::LIGHTREMOTE);
	key.enable();

	ngl::Light fill(ngl::Vector(2,0.5,1),ngl::Colour(0.6,0.6,0.6,1),ngl::LIGHTLOCAL);
	fill.enable();

	ngl::Light back(ngl::Vector(0,1,-1),ngl::Colour(0.8,0.8,0.8,1),ngl::LIGHTLOCAL);
	back.enable();

}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{

  glViewport(0,0,_w,_h);
  m_cam->setShape(25,(float)_w/_h,0.0001,350,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["TextureShader"]->use();
	shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
}

// a simple structure to hold our vertex data
struct vertData
{
  GLfloat u; // tex cords from obj
	GLfloat v; // tex cords
	GLfloat nx; // normal from obj mesh
	GLfloat ny;
	GLfloat nz;
	GLfloat x; // position from obj
	GLfloat y;
	GLfloat z;
	GLfloat tx; // tangent calculated by us
	GLfloat ty;
	GLfloat tz;
	GLfloat bx; // binormal (bi-tangent really) calculated by us
	GLfloat by;
	GLfloat bz;
};


void GLWindow::loadModel()
{
	std::cout<<"loading mesh\n";
	// load the obj
	ngl::Obj mesh("models/troll.obj");
	std::cout<<"checking triangular\n";
	// this is only going to work for tri meshes so check
	if( ! mesh.isTriangular() )
	{
		std::cout<<"only works for tri meshes\n";
		exit(EXIT_FAILURE);
	}
	std::cout<<"getting lists\n";

	// get the obj data so we can process it locally
	std::vector <ngl::Vector> verts=mesh.getVertexList();
	std::vector <ngl::Face> faces=mesh.getFaceList();
	std::vector <ngl::Vector> tex=mesh.getTextureCordList();
	std::vector <ngl::Vector> normals=mesh.getNormalList();
	std::cout<<"got mesh data\n";
	// now we are going to process and pack the mesh into an ngl::VertexArrayObject
	std::vector <vertData> vboMesh;
	vertData d;
	unsigned int nFaces=faces.size();
	unsigned int nNorm=normals.size();
	unsigned int nTex=tex.size();
	// loop for each of the faces
	for(unsigned int i=0;i<nFaces;++i)
	{
		// now for each triangle in the face (remember we ensured tri above)
		for(int j=0;j<3;++j)
		{
			// pack in the vertex data first
			d.x=verts[faces[i].m_vert[j]].m_x;
			d.y=verts[faces[i].m_vert[j]].m_y;
			d.z=verts[faces[i].m_vert[j]].m_z;
			// now if we have norms of tex (possibly could not) pack them as well
      if(nNorm >0 && nTex > 0)
      {
        d.nx=normals[faces[i].m_norm[j]].m_x;
        d.ny=normals[faces[i].m_norm[j]].m_y;
        d.nz=normals[faces[i].m_norm[j]].m_z;

				d.u=tex[faces[i].m_tex[j]].m_x;
				d.v=tex[faces[i].m_tex[j]].m_y;

      }
      // now if neither are present (only verts like Zbrush models)
      else if(nNorm ==0 && nTex==0)
      {
        d.nx=0;
        d.ny=0;
        d.nz=0;
				d.u=0;
				d.v=0;
      }
      // here we've got norms but not tex
      else if(nNorm >0 && nTex==0)
      {
        d.nx=normals[faces[i].m_norm[j]].m_x;
        d.ny=normals[faces[i].m_norm[j]].m_y;
        d.nz=normals[faces[i].m_norm[j]].m_z;
				d.u=0;
				d.v=0;

      }
      // here we've got tex but not norm least common
      else if(nNorm ==0 && nTex>0)
			{
        d.nx=0;
        d.ny=0;
        d.nz=0;
				d.u=tex[faces[i].m_tex[j]].m_x;
				d.v=tex[faces[i].m_tex[j]].m_y;
			}
		// now we calculate the tangent / bi-normal (tangent) based on the article here
		// http://www.terathon.com/code/tangent.html

		ngl::Vector c1 = normals[faces[i].m_norm[j]].cross(ngl::Vector(0.0, 0.0, 1.0));
		ngl::Vector c2 = normals[faces[i].m_norm[j]].cross(ngl::Vector(0.0, 1.0, 0.0));
		ngl::Vector tangent;
		ngl::Vector binormal;
		if(c1.length()>c2.length())
		{
			tangent = c1;
		}
		else
		{
			tangent = c2;
		}
		// now we normalize the tangent so we don't need to do it in the shader
		tangent.normalize();
		// now we calculate the binormal using the model normal and tangent (cross)
		binormal = normals[faces[i].m_norm[j]].cross(tangent);
		// normalize again so we don't need to in the shader
		binormal.normalize();
		d.tx=tangent.m_x;
		d.ty=tangent.m_y;
		d.tz=tangent.m_z;
		d.bx=binormal.m_x;
		d.by=binormal.m_y;
		d.bz=binormal.m_z;
		// finally add it to our mesh VAO structure
		vboMesh.push_back(d);
		}
	}

	// first we grab an instance of our VOA class as a TRIANGLE_STRIP
	m_vaoMesh= ngl::VertexArrayObject::createVOA(GL_TRIANGLES);
	// next we bind it so it's active for setting data
	m_vaoMesh->bind();
	unsigned int meshSize=vboMesh.size();
	// now we have our data add it to the VAO, we need to tell the VAO the following
	// how much (in bytes) data we are copying
	// a pointer to the first element of data (in this case the address of the first element of the
	// std::vector
	m_vaoMesh->setData(meshSize*sizeof(vertData),vboMesh[0].u);
	// in this case we have packed our data in interleaved format as follows
	// u,v,nx,ny,nz,x,y,z
	// If you look at the shader we have the following attributes being used
	// attribute vec3 inVert; attribute 0
	// attribute vec2 inUV; attribute 1
	// attribute vec3 inNormal; attribure 2
	// so we need to set the vertexAttributePointer so the correct size and type as follows
	// vertex is attribute 0 with x,y,z(3) parts of type GL_FLOAT, our complete packed data is
	// sizeof(vertData) and the offset into the data structure for the first x component is 5 (u,v,nx,ny,nz)..x
	m_vaoMesh->setVertexAttributePointer(0,3,GL_FLOAT,sizeof(vertData),5);
	// uv same as above but starts at 0 and is attrib 1 and only u,v so 2
	m_vaoMesh->setVertexAttributePointer(1,2,GL_FLOAT,sizeof(vertData),0);
	// normal same as vertex only starts at position 2 (u,v)-> nx
	m_vaoMesh->setVertexAttributePointer(2,3,GL_FLOAT,sizeof(vertData),2);

	// tangent same as vertex only starts at position 8 (u,v)-> nx
	m_vaoMesh->setVertexAttributePointer(3,3,GL_FLOAT,sizeof(vertData),8);

	// bi-tangent (or Binormal) same as vertex only starts at position 11 (u,v)-> nx
	m_vaoMesh->setVertexAttributePointer(4,3,GL_FLOAT,sizeof(vertData),11);


	// now we have set the vertex attributes we tell the VAO class how many indices to draw when
	// glDrawArrays is called, in this case we use buffSize (but if we wished less of the sphere to be drawn we could
	// specify less (in steps of 3))
	m_vaoMesh->setNumIndices(meshSize);
	// finally we have finished for now so time to unbind the VAO
	m_vaoMesh->unbind();

}





//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
	// load the m_modePos into the global transform, this will move the model
	// based on the mouse / wheel movements
	trans.setPosition(m_modelPos);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);
	// activate the shader
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["TextureShader"]->use();
	// load TX stack into shader
  m_transformStack.loadGlobalAndCurrentMatrixToShader("TextureShader","ModelMatrix");

	shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	// now we bind back our vertex array object and draw
	m_vaoMesh->bind();
	// draw the VAO
	m_vaoMesh->draw();

	// now we are done so unbind
	m_vaoMesh->unbind();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
	// right mouse translate code
  else if(m_translate && _event->buttons() == Qt::RightButton)
  {
		int diffX = (int)(_event->x() - m_origXPos);
		int diffY = (int)(_event->y() - m_origYPos);
		m_origXPos=_event->x();
    m_origYPos=_event->y();
		m_modelPos.m_x += INCREMENT * diffX;
		m_modelPos.m_y -= INCREMENT * diffY;

	}
  updateGL();

}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
	// right mouse translate mode
	else if(_event->button() == Qt::RightButton)
	{
    m_origXPos = _event->x();
    m_origYPos = _event->y();
		m_translate=true;
	}
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
	// right mouse translate mode
  if (_event->button() == Qt::RightButton)
  {
    m_translate=false;
  }
}

void GLWindow::wheelEvent(QWheelEvent *_event)
{

	// check the diff of the wheel position (0 means no change)
	if(_event->delta() > 0)
	{
		m_modelPos.m_z+=ZOOM;
	}
	else if(_event->delta() <0 )
	{
		m_modelPos.m_z-=ZOOM;
	}
	updateGL();
}

